@@ -45,15 +45,15 @@
 
 QT_BEGIN_NAMESPACE
 
-QQmlPropertyValidator::QQmlPropertyValidator(QQmlEnginePrivate *enginePrivate, const QQmlImports &imports, QV4::CompiledData::CompilationUnit *compilationUnit)
+QQmlPropertyValidator::QQmlPropertyValidator(QQmlEnginePrivate *enginePrivate, const QQmlImports &imports, const QQmlRefPointer<QV4::CompiledData::CompilationUnit> &compilationUnit)
     : enginePrivate(enginePrivate)
+    , compilationUnit(compilationUnit)
     , imports(imports)
-    , qmlUnit(compilationUnit->data)
-    , resolvedTypes(compilationUnit->resolvedTypes)
+    , qmlUnit(compilationUnit->unitData())
     , propertyCaches(compilationUnit->propertyCaches)
     , bindingPropertyDataPerObject(&compilationUnit->bindingPropertyDataPerObject)
 {
-    bindingPropertyDataPerObject->resize(qmlUnit->nObjects);
+    bindingPropertyDataPerObject->resize(compilationUnit->objectCount());
 }
 
 QVector<QQmlCompileError> QQmlPropertyValidator::validate()
@@ -81,7 +81,7 @@ struct BindingFinder
 
 QVector<QQmlCompileError> QQmlPropertyValidator::validateObject(int objectIndex, const QV4::CompiledData::Binding *instantiatingBinding, bool populatingValueTypeGroupProperty) const
 {
-    const QV4::CompiledData::Object *obj = qmlUnit->objectAt(objectIndex);
+    const QV4::CompiledData::Object *obj = compilationUnit->objectAt(objectIndex);
 
     if (obj->flags & QV4::CompiledData::Object::IsComponent) {
         Q_ASSERT(obj->nBindings == 1);
@@ -95,7 +95,7 @@ QVector<QQmlCompileError> QQmlPropertyValidator::validateObject(int objectIndex,
         return QVector<QQmlCompileError>();
 
     QQmlCustomParser *customParser = nullptr;
-    if (auto typeRef = resolvedTypes.value(obj->inheritedTypeNameIndex)) {
+    if (auto typeRef = resolvedType(obj->inheritedTypeNameIndex)) {
         if (typeRef->type.isValid())
             customParser = typeRef->type.customParser();
     }
@@ -167,7 +167,7 @@ QVector<QQmlCompileError> QQmlPropertyValidator::validateObject(int objectIndex,
 
             if (notInRevision) {
                 QString typeName = stringAt(obj->inheritedTypeNameIndex);
-                auto *objectType = resolvedTypes.value(obj->inheritedTypeNameIndex);
+                auto *objectType = resolvedType(obj->inheritedTypeNameIndex);
                 if (objectType && objectType->type.isValid()) {
                     return recordError(binding->location, tr("\"%1.%2\" is not available in %3 %4.%5.").arg(typeName).arg(name).arg(objectType->type.module()).arg(objectType->majorVersion).arg(objectType->minorVersion));
                 } else {
@@ -275,7 +275,11 @@ QVector<QQmlCompileError> QQmlPropertyValidator::validateObject(int objectIndex,
                     }
                 } else {
                     if (!enginePrivate->propertyCacheForType(pd->propType())) {
-                        return recordError(binding->location, tr("Invalid grouped property access"));
+                        return recordError(binding->location,
+                                           tr("Invalid grouped property access: Property \"%1\" with type \"%2\", which is not a value type")
+                                           .arg(name)
+                                           .arg(QString::fromLatin1(QMetaType::typeName(pd->propType())))
+                                          );
                     }
                 }
             }
@@ -305,7 +309,7 @@ QVector<QQmlCompileError> QQmlPropertyValidator::validateObject(int objectIndex,
         customParser->validator = this;
         customParser->engine = enginePrivate;
         customParser->imports = &imports;
-        customParser->verifyBindings(qmlUnit, customBindings);
+        customParser->verifyBindings(compilationUnit, customBindings);
         customParser->validator = nullptr;
         customParser->engine = nullptr;
         customParser->imports = (QQmlImports*)nullptr;
@@ -332,7 +336,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
         if (binding->flags & QV4::CompiledData::Binding::IsResolvedEnum)
             return noError;
 
-        QString value = binding->valueAsString(qmlUnit);
+        QString value = binding->valueAsString(compilationUnit.data());
         QMetaProperty p = propertyCache->firstCppMetaObject()->property(property->coreIndex());
         bool ok;
         if (p.isFlagType()) {
@@ -375,7 +379,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::UInt: {
         if (binding->type == QV4::CompiledData::Binding::Type_Number) {
-            double d = binding->valueAsNumber();
+            double d = binding->valueAsNumber(compilationUnit->constants);
             if (double(uint(d)) == d)
                 return noError;
         }
@@ -384,7 +388,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::Int: {
         if (binding->type == QV4::CompiledData::Binding::Type_Number) {
-            double d = binding->valueAsNumber();
+            double d = binding->valueAsNumber(compilationUnit->constants);
             if (double(int(d)) == d)
                 return noError;
         }
@@ -405,7 +409,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::Color: {
         bool ok = false;
-        QQmlStringConverters::rgbaFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::rgbaFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: color expected"));
         }
@@ -414,7 +418,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
 #if QT_CONFIG(datestring)
     case QVariant::Date: {
         bool ok = false;
-        QQmlStringConverters::dateFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::dateFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: date expected"));
         }
@@ -422,7 +426,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::Time: {
         bool ok = false;
-        QQmlStringConverters::timeFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::timeFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: time expected"));
         }
@@ -430,7 +434,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::DateTime: {
         bool ok = false;
-        QQmlStringConverters::dateTimeFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::dateTimeFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: datetime expected"));
         }
@@ -439,7 +443,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
 #endif // datestring
     case QVariant::Point: {
         bool ok = false;
-        QQmlStringConverters::pointFFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::pointFFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: point expected"));
         }
@@ -447,7 +451,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::PointF: {
         bool ok = false;
-        QQmlStringConverters::pointFFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::pointFFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: point expected"));
         }
@@ -455,7 +459,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::Size: {
         bool ok = false;
-        QQmlStringConverters::sizeFFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::sizeFFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: size expected"));
         }
@@ -463,7 +467,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::SizeF: {
         bool ok = false;
-        QQmlStringConverters::sizeFFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::sizeFFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: size expected"));
         }
@@ -471,7 +475,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::Rect: {
         bool ok = false;
-        QQmlStringConverters::rectFFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::rectFFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: rect expected"));
         }
@@ -479,7 +483,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
     break;
     case QVariant::RectF: {
         bool ok = false;
-        QQmlStringConverters::rectFFromString(binding->valueAsString(qmlUnit), &ok);
+        QQmlStringConverters::rectFFromString(binding->valueAsString(compilationUnit.data()), &ok);
         if (!ok) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: point expected"));
         }
@@ -496,7 +500,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
             float xp;
             float yp;
         } vec;
-        if (!QQmlStringConverters::createFromString(QMetaType::QVector2D, binding->valueAsString(qmlUnit), &vec, sizeof(vec))) {
+        if (!QQmlStringConverters::createFromString(QMetaType::QVector2D, binding->valueAsString(compilationUnit.data()), &vec, sizeof(vec))) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: 2D vector expected"));
         }
     }
@@ -507,7 +511,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
             float yp;
             float zy;
         } vec;
-        if (!QQmlStringConverters::createFromString(QMetaType::QVector3D, binding->valueAsString(qmlUnit), &vec, sizeof(vec))) {
+        if (!QQmlStringConverters::createFromString(QMetaType::QVector3D, binding->valueAsString(compilationUnit.data()), &vec, sizeof(vec))) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: 3D vector expected"));
         }
     }
@@ -519,7 +523,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
             float zy;
             float wp;
         } vec;
-        if (!QQmlStringConverters::createFromString(QMetaType::QVector4D, binding->valueAsString(qmlUnit), &vec, sizeof(vec))) {
+        if (!QQmlStringConverters::createFromString(QMetaType::QVector4D, binding->valueAsString(compilationUnit.data()), &vec, sizeof(vec))) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: 4D vector expected"));
         }
     }
@@ -531,7 +535,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
             float yp;
             float zp;
         } vec;
-        if (!QQmlStringConverters::createFromString(QMetaType::QQuaternion, binding->valueAsString(qmlUnit), &vec, sizeof(vec))) {
+        if (!QQmlStringConverters::createFromString(QMetaType::QQuaternion, binding->valueAsString(compilationUnit.data()), &vec, sizeof(vec))) {
             return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: quaternion expected"));
         }
     }
@@ -548,7 +552,7 @@ QQmlCompileError QQmlPropertyValidator::validateLiteralBinding(QQmlPropertyCache
         } else if (property->propType() == qMetaTypeId<QList<int> >()) {
             bool ok = (binding->type == QV4::CompiledData::Binding::Type_Number);
             if (ok) {
-                double n = binding->valueAsNumber();
+                double n = binding->valueAsNumber(compilationUnit->constants);
                 if (double(int(n)) != n)
                     ok = false;
             }
@@ -627,9 +631,9 @@ QQmlCompileError QQmlPropertyValidator::validateObjectBinding(QQmlPropertyData *
         bool isValueSource = false;
         bool isPropertyInterceptor = false;
 
-        const QV4::CompiledData::Object *targetObject = qmlUnit->objectAt(binding->value.objectIndex);
-        if (auto *typeRef = resolvedTypes.value(targetObject->inheritedTypeNameIndex)) {
-            QQmlPropertyCache *cache = typeRef->createPropertyCache(QQmlEnginePrivate::get(enginePrivate));
+        const QV4::CompiledData::Object *targetObject = compilationUnit->objectAt(binding->value.objectIndex);
+        if (auto *typeRef = resolvedType(targetObject->inheritedTypeNameIndex)) {
+            QQmlRefPointer<QQmlPropertyCache> cache = typeRef->createPropertyCache(QQmlEnginePrivate::get(enginePrivate));
             const QMetaObject *mo = cache->firstCppMetaObject();
             QQmlType qmlType;
             while (mo && !qmlType.isValid()) {
@@ -665,12 +669,12 @@ QQmlCompileError QQmlPropertyValidator::validateObjectBinding(QQmlPropertyData *
             }
         }
         return noError;
-    } else if (qmlUnit->objectAt(binding->value.objectIndex)->flags & QV4::CompiledData::Object::IsComponent) {
+    } else if (compilationUnit->objectAt(binding->value.objectIndex)->flags & QV4::CompiledData::Object::IsComponent) {
         return noError;
     } else if (binding->flags & QV4::CompiledData::Binding::IsSignalHandlerObject && property->isFunction()) {
         return noError;
     } else if (QQmlValueTypeFactory::isValueType(property->propType())) {
-        return QQmlCompileError(binding->location, tr("Unexpected object assignment"));
+        return QQmlCompileError(binding->location, tr("Unexpected object assignment for property \"%1\"").arg(propertyName));
     } else if (property->propType() == qMetaTypeId<QQmlScriptString>()) {
         return QQmlCompileError(binding->valueLocation, tr("Invalid property assignment: script expected"));
     } else {
@@ -680,7 +684,7 @@ QQmlCompileError QQmlPropertyValidator::validateObjectBinding(QQmlPropertyData *
         // Using -1 for the minor version ensures that we get the raw metaObject.
         QQmlPropertyCache *propertyMetaObject = enginePrivate->rawPropertyCacheForType(property->propType(), -1);
 
-        // Will be true if the assgned type inherits propertyMetaObject
+        // Will be true if the assigned type inherits propertyMetaObject
         bool isAssignable = false;
         // Determine isAssignable value
         if (propertyMetaObject) {
@@ -692,7 +696,8 @@ QQmlCompileError QQmlPropertyValidator::validateObjectBinding(QQmlPropertyData *
         }
 
         if (!isAssignable) {
-            return QQmlCompileError(binding->valueLocation, tr("Cannot assign object to property"));
+            return QQmlCompileError(binding->valueLocation, tr("Cannot assign object of type \"%1\" to property of type \"%2\" as the former is neither the same as the latter nor a sub-class of it.")
+                    .arg(stringAt(compilationUnit->objectAt(binding->value.objectIndex)->inheritedTypeNameIndex)).arg(QLatin1String(QMetaType::typeName(property->propType()))));
         }
     }
     return noError;
