@@ -47,13 +47,13 @@ QT_BEGIN_NAMESPACE
 using namespace QV4;
 
 
-void Lookup::resolveProtoGetter(Identifier *name, const Heap::Object *proto)
+void Lookup::resolveProtoGetter(PropertyKey name, const Heap::Object *proto)
 {
     while (proto) {
-        uint index = proto->internalClass->find(name);
-        if (index != UINT_MAX) {
-            PropertyAttributes attrs = proto->internalClass->propertyData.at(index);
-            protoLookup.data = proto->propertyData(index);
+        auto index = proto->internalClass->findValueOrGetter(name);
+        if (index.isValid()) {
+            PropertyAttributes attrs = index.attrs;
+            protoLookup.data = proto->propertyData(index.index);
             if (attrs.isData()) {
                 getter = getterProto;
             } else {
@@ -70,29 +70,34 @@ void Lookup::resolveProtoGetter(Identifier *name, const Heap::Object *proto)
 ReturnedValue Lookup::resolveGetter(ExecutionEngine *engine, const Object *object)
 {
     Heap::Object *obj = object->d();
-    Identifier *name = engine->identifierTable->identifier(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[nameIndex]);
+    PropertyKey name = engine->identifierTable->asPropertyKey(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[nameIndex]);
+    if (name.isArrayIndex()) {
+        indexedLookup.index = name.asArrayIndex();
+        getter = getterIndexed;
+        return getter(this, engine, *object);
+    }
 
-    uint index = obj->internalClass->find(name);
-    if (index != UINT_MAX) {
-        PropertyAttributes attrs = obj->internalClass->propertyData.at(index);
+    auto index = obj->internalClass->findValueOrGetter(name);
+    if (index.isValid()) {
+        PropertyAttributes attrs = index.attrs;
         uint nInline = obj->vtable()->nInlineProperties;
         if (attrs.isData()) {
-            if (index < obj->vtable()->nInlineProperties) {
-                index += obj->vtable()->inlinePropertyOffset;
+            if (index.index < obj->vtable()->nInlineProperties) {
+                index.index += obj->vtable()->inlinePropertyOffset;
                 getter = getter0Inline;
             } else {
-                index -= nInline;
+                index.index -= nInline;
                 getter = getter0MemberData;
             }
         } else {
             getter = getterAccessor;
         }
         objectLookup.ic = obj->internalClass;
-        objectLookup.offset = index;
+        objectLookup.offset = index.index;
         return getter(this, engine, *object);
     }
 
-    protoLookup.icIdentifier = obj->internalClass->id;
+    protoLookup.protoId = obj->internalClass->protoId;
     resolveProtoGetter(name, obj->prototype());
     return getter(this, engine, *object);
 }
@@ -109,11 +114,12 @@ ReturnedValue Lookup::resolvePrimitiveGetter(ExecutionEngine *engine, const Valu
         break;
     case Value::Managed_Type: {
         // ### Should move this over to the Object path, as strings also have an internalClass
-        Q_ASSERT(object.isString());
-        primitiveLookup.proto = engine->stringPrototype()->d();
+        Q_ASSERT(object.isStringOrSymbol());
+        primitiveLookup.proto = static_cast<const Managed &>(object).internalClass()->prototype;
+        Q_ASSERT(primitiveLookup.proto);
         Scope scope(engine);
         ScopedString name(scope, engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[nameIndex]);
-        if (name->equals(engine->id_length())) {
+        if (object.isString() && name->equals(engine->id_length())) {
             // special case, as the property is on the object itself
             getter = stringLengthGetter;
             return stringLengthGetter(this, engine, object);
@@ -125,8 +131,8 @@ ReturnedValue Lookup::resolvePrimitiveGetter(ExecutionEngine *engine, const Valu
         primitiveLookup.proto = engine->numberPrototype()->d();
     }
 
-    Identifier *name = engine->identifierTable->identifier(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[nameIndex]);
-    protoLookup.icIdentifier = primitiveLookup.proto->internalClass->id;
+    PropertyKey name = engine->identifierTable->asPropertyKey(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[nameIndex]);
+    protoLookup.protoId = primitiveLookup.proto->internalClass->protoId;
     resolveProtoGetter(name, primitiveLookup.proto);
 
     if (getter == getterProto)
@@ -139,8 +145,8 @@ ReturnedValue Lookup::resolvePrimitiveGetter(ExecutionEngine *engine, const Valu
 ReturnedValue Lookup::resolveGlobalGetter(ExecutionEngine *engine)
 {
     Object *o = engine->globalObject;
-    Identifier *name = engine->identifierTable->identifier(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[nameIndex]);
-    protoLookup.icIdentifier = o->internalClass()->id;
+    PropertyKey name = engine->identifierTable->asPropertyKey(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[nameIndex]);
+    protoLookup.protoId = o->internalClass()->protoId;
     resolveProtoGetter(name, o->d());
 
     if (getter == getterProto)
@@ -188,16 +194,16 @@ ReturnedValue Lookup::getterTwoClasses(Lookup *l, ExecutionEngine *engine, const
             return result;
         }
         if (first.getter == getterProto && second.getter == getterProto) {
-            l->protoLookupTwoClasses.icIdentifier = first.protoLookup.icIdentifier;
-            l->protoLookupTwoClasses.icIdentifier2 = second.protoLookup.icIdentifier;
+            l->protoLookupTwoClasses.protoId = first.protoLookup.protoId;
+            l->protoLookupTwoClasses.protoId2 = second.protoLookup.protoId;
             l->protoLookupTwoClasses.data = first.protoLookup.data;
             l->protoLookupTwoClasses.data2 = second.protoLookup.data;
             l->getter = getterProtoTwoClasses;
             return result;
         }
         if (first.getter == getterProtoAccessor && second.getter == getterProtoAccessor) {
-            l->protoLookupTwoClasses.icIdentifier = first.protoLookup.icIdentifier;
-            l->protoLookupTwoClasses.icIdentifier2 = second.protoLookup.icIdentifier;
+            l->protoLookupTwoClasses.protoId = first.protoLookup.protoId;
+            l->protoLookupTwoClasses.protoId2 = second.protoLookup.protoId;
             l->protoLookupTwoClasses.data = first.protoLookup.data;
             l->protoLookupTwoClasses.data2 = second.protoLookup.data;
             l->getter = getterProtoAccessorTwoClasses;
@@ -250,7 +256,7 @@ ReturnedValue Lookup::getterProto(Lookup *l, ExecutionEngine *engine, const Valu
     // the internal class won't match
     Heap::Object *o = static_cast<Heap::Object *>(object.heapObject());
     if (o) {
-        if (l->protoLookup.icIdentifier == o->internalClass->id)
+        if (l->protoLookup.protoId == o->internalClass->protoId)
             return l->protoLookup.data->asReturnedValue();
     }
     return getterTwoClasses(l, engine, object);
@@ -266,6 +272,9 @@ ReturnedValue Lookup::getter0Inlinegetter0Inline(Lookup *l, ExecutionEngine *eng
             return o->inlinePropertyDataWithOffset(l->objectLookupTwoClasses.offset)->asReturnedValue();
         if (l->objectLookupTwoClasses.ic2 == o->internalClass)
             return o->inlinePropertyDataWithOffset(l->objectLookupTwoClasses.offset2)->asReturnedValue();
+        Value obj = Value::fromHeapObject(o);
+        Value str = Value::fromHeapObject(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[l->nameIndex]);
+        return static_cast<Object &>(obj).get(&static_cast<String &>(str));
     }
     l->getter = getterFallback;
     return getterFallback(l, engine, object);
@@ -281,6 +290,9 @@ ReturnedValue Lookup::getter0Inlinegetter0MemberData(Lookup *l, ExecutionEngine
             return o->inlinePropertyDataWithOffset(l->objectLookupTwoClasses.offset)->asReturnedValue();
         if (l->objectLookupTwoClasses.ic2 == o->internalClass)
             return o->memberData->values.data()[l->objectLookupTwoClasses.offset2].asReturnedValue();
+        Value obj = Value::fromHeapObject(o);
+        Value str = Value::fromHeapObject(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[l->nameIndex]);
+        return static_cast<Object &>(obj).get(&static_cast<String &>(str));
     }
     l->getter = getterFallback;
     return getterFallback(l, engine, object);
@@ -296,6 +308,9 @@ ReturnedValue Lookup::getter0MemberDatagetter0MemberData(Lookup *l, ExecutionEng
             return o->memberData->values.data()[l->objectLookupTwoClasses.offset].asReturnedValue();
         if (l->objectLookupTwoClasses.ic2 == o->internalClass)
             return o->memberData->values.data()[l->objectLookupTwoClasses.offset2].asReturnedValue();
+        Value obj = Value::fromHeapObject(o);
+        Value str = Value::fromHeapObject(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[l->nameIndex]);
+        return static_cast<Object &>(obj).get(&static_cast<String &>(str));
     }
     l->getter = getterFallback;
     return getterFallback(l, engine, object);
@@ -307,11 +322,13 @@ ReturnedValue Lookup::getterProtoTwoClasses(Lookup *l, ExecutionEngine *engine,
     // the internal class won't match
     Heap::Object *o = static_cast<Heap::Object *>(object.heapObject());
     if (o) {
-        if (l->protoLookupTwoClasses.icIdentifier == o->internalClass->id)
+        if (l->protoLookupTwoClasses.protoId == o->internalClass->protoId)
             return l->protoLookupTwoClasses.data->asReturnedValue();
-        if (l->protoLookupTwoClasses.icIdentifier2 == o->internalClass->id)
+        if (l->protoLookupTwoClasses.protoId2 == o->internalClass->protoId)
             return l->protoLookupTwoClasses.data2->asReturnedValue();
-        return getterFallback(l, engine, object);
+        Value obj = Value::fromHeapObject(o);
+        Value str = Value::fromHeapObject(engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[l->nameIndex]);
+        return static_cast<Object &>(obj).get(&static_cast<String &>(str));
     }
     l->getter = getterFallback;
     return getterFallback(l, engine, object);
@@ -340,14 +357,13 @@ ReturnedValue Lookup::getterProtoAccessor(Lookup *l, ExecutionEngine *engine, co
     // we can safely cast to a QV4::Object here. If object is actually a string,
     // the internal class won't match
     Heap::Object *o = static_cast<Heap::Object *>(object.heapObject());
-    if (o && l->protoLookup.icIdentifier == o->internalClass->id) {
+    if (o && l->protoLookup.protoId == o->internalClass->protoId) {
         const Value *getter = l->protoLookup.data;
         if (!getter->isFunctionObject()) // ### catch at resolve time
             return Encode::undefined();
 
         return static_cast<const FunctionObject *>(getter)->call(&object, nullptr, 0);
     }
-    l->getter = getterTwoClasses;
     return getterTwoClasses(l, engine, object);
 }
 
@@ -358,9 +374,9 @@ ReturnedValue Lookup::getterProtoAccessorTwoClasses(Lookup *l, ExecutionEngine *
     Heap::Object *o = static_cast<Heap::Object *>(object.heapObject());
     if (o) {
         const Value *getter = nullptr;
-        if (l->protoLookupTwoClasses.icIdentifier == o->internalClass->id)
+        if (l->protoLookupTwoClasses.protoId == o->internalClass->protoId)
             getter = l->protoLookupTwoClasses.data;
-        else if (l->protoLookupTwoClasses.icIdentifier2 == o->internalClass->id)
+        else if (l->protoLookupTwoClasses.protoId2 == o->internalClass->protoId)
             getter = l->protoLookupTwoClasses.data2;
         if (getter) {
             if (!getter->isFunctionObject()) // ### catch at resolve time
@@ -373,11 +389,29 @@ ReturnedValue Lookup::getterProtoAccessorTwoClasses(Lookup *l, ExecutionEngine *
     return getterFallback(l, engine, object);
 }
 
+ReturnedValue Lookup::getterIndexed(Lookup *l, ExecutionEngine *engine, const Value &object)
+{
+    Object *o = object.objectValue();
+    if (o) {
+        Heap::Object *ho = o->d();
+        if (ho->arrayData && ho->arrayData->type == Heap::ArrayData::Simple) {
+            Heap::SimpleArrayData *s = ho->arrayData.cast<Heap::SimpleArrayData>();
+            if (l->indexedLookup.index < s->values.size)
+                if (!s->data(l->indexedLookup.index).isEmpty())
+                    return s->data(l->indexedLookup.index).asReturnedValue();
+        }
+        return o->get(l->indexedLookup.index);
+    }
+    l->getter = getterFallback;
+    return getterFallback(l, engine, object);
+
+}
+
 ReturnedValue Lookup::primitiveGetterProto(Lookup *l, ExecutionEngine *engine, const Value &object)
 {
     if (object.type() == l->primitiveLookup.type) {
         Heap::Object *o = l->primitiveLookup.proto;
-        if (l->primitiveLookup.icIdentifier == o->internalClass->id)
+        if (l->primitiveLookup.protoId == o->internalClass->protoId)
             return l->primitiveLookup.data->asReturnedValue();
     }
     l->getter = getterGeneric;
@@ -388,7 +422,7 @@ ReturnedValue Lookup::primitiveGetterAccessor(Lookup *l, ExecutionEngine *engine
 {
     if (object.type() == l->primitiveLookup.type) {
         Heap::Object *o = l->primitiveLookup.proto;
-        if (l->primitiveLookup.icIdentifier == o->internalClass->id) {
+        if (l->primitiveLookup.protoId == o->internalClass->protoId) {
             const Value *getter = l->primitiveLookup.data;
             if (!getter->isFunctionObject()) // ### catch at resolve time
                 return Encode::undefined();
@@ -417,7 +451,7 @@ ReturnedValue Lookup::globalGetterGeneric(Lookup *l, ExecutionEngine *engine)
 ReturnedValue Lookup::globalGetterProto(Lookup *l, ExecutionEngine *engine)
 {
     Heap::Object *o = engine->globalObject->d();
-    if (l->protoLookup.icIdentifier == o->internalClass->id)
+    if (l->protoLookup.protoId == o->internalClass->protoId)
         return l->protoLookup.data->asReturnedValue();
     l->globalGetter = globalGetterGeneric;
     return globalGetterGeneric(l, engine);
@@ -426,7 +460,7 @@ ReturnedValue Lookup::globalGetterProto(Lookup *l, ExecutionEngine *engine)
 ReturnedValue Lookup::globalGetterProtoAccessor(Lookup *l, ExecutionEngine *engine)
 {
     Heap::Object *o = engine->globalObject->d();
-    if (l->protoLookup.icIdentifier == o->internalClass->id) {
+    if (l->protoLookup.protoId == o->internalClass->protoId) {
         const Value *getter = l->protoLookup.data;
         if (!getter->isFunctionObject()) // ### catch at resolve time
             return Encode::undefined();
@@ -442,16 +476,18 @@ bool Lookup::resolveSetter(ExecutionEngine *engine, Object *object, const Value
     Scope scope(engine);
     ScopedString name(scope, scope.engine->currentStackFrame->v4Function->compilationUnit->runtimeStrings[nameIndex]);
 
-    InternalClass *c = object->internalClass();
-    uint idx = c->find(name);
-    if (idx != UINT_MAX) {
-        if (object->isArrayObject() && idx == Heap::ArrayObject::LengthPropertyIndex) {
+    Heap::InternalClass *c = object->internalClass();
+    PropertyKey key = name->toPropertyKey();
+    auto idx = c->findValueOrSetter(key);
+    if (idx.isValid()) {
+        if (object->isArrayObject() && idx.index == Heap::ArrayObject::LengthPropertyIndex) {
+            Q_ASSERT(!idx.attrs.isAccessor());
             setter = arrayLengthSetter;
             return setter(this, engine, *object, value);
-        } else if (object->internalClass()->propertyData[idx].isData() && object->internalClass()->propertyData[idx].isWritable()) {
+        } else if (idx.attrs.isData() && idx.attrs.isWritable()) {
             objectLookup.ic = object->internalClass();
-            objectLookup.offset = idx;
-            setter = idx < object->d()->vtable()->nInlineProperties ? Lookup::setter0Inline : Lookup::setter0;
+            objectLookup.offset = idx.index;
+            setter = idx.index < object->d()->vtable()->nInlineProperties ? Lookup::setter0Inline : Lookup::setter0;
             return setter(this, engine, *object, value);
         } else {
             // ### handle setter
@@ -460,8 +496,8 @@ bool Lookup::resolveSetter(ExecutionEngine *engine, Object *object, const Value
         return setter(this, engine, *object, value);
     }
 
-    insertionLookup.icIdentifier = c->id;
-    if (!object->put(name, value)) {
+    insertionLookup.protoId = c->protoId;
+    if (!object->put(key, value)) {
         setter = Lookup::setterFallback;
         return false;
     }
@@ -471,13 +507,13 @@ bool Lookup::resolveSetter(ExecutionEngine *engine, Object *object, const Value
         setter = setterFallback;
         return true;
     }
-    idx = object->internalClass()->find(name);
-    if (idx == UINT_MAX) { // ### can this even happen?
+    idx = object->internalClass()->findValueOrSetter(key);
+    if (!idx.isValid() || idx.attrs.isAccessor()) { // ### can this even happen?
         setter = setterFallback;
         return false;
     }
     insertionLookup.newClass = object->internalClass();
-    insertionLookup.offset = idx;
+    insertionLookup.offset = idx.index;
     setter = setterInsert;
     return true;
 }
@@ -487,6 +523,9 @@ bool Lookup::setterGeneric(Lookup *l, ExecutionEngine *engine, Value &object, co
     if (object.isObject())
         return l->resolveSetter(engine, static_cast<Object *>(&object), value);
 
+    if (engine->currentStackFrame->v4Function->isStrict())
+        return false;
+
     Scope scope(engine);
     ScopedObject o(scope, RuntimeHelpers::convertToObject(scope.engine, object));
     if (!o) // type error
@@ -574,7 +613,7 @@ bool Lookup::setter0setter0(Lookup *l, ExecutionEngine *engine, Value &object, c
 bool Lookup::setterInsert(Lookup *l, ExecutionEngine *engine, Value &object, const Value &value)
 {
     Object *o = static_cast<Object *>(object.managed());
-    if (o && o->internalClass()->id == l->insertionLookup.icIdentifier) {
+    if (o && o->internalClass()->protoId == l->insertionLookup.protoId) {
         o->setInternalClass(l->insertionLookup.newClass);
         o->d()->setProperty(engine, l->insertionLookup.offset, value);
         return true;
