@@ -48,7 +48,6 @@
 #include "qqmlcomponent.h"
 #include "qqmlvme_p.h"
 #include "qqmlstringconverters_p.h"
-#include "qqmlxmlhttprequest_p.h"
 #include "qqmlscriptstring.h"
 #include "qqmlglobal_p.h"
 #include "qqmlcomponent_p.h"
@@ -79,20 +78,26 @@
 
 #include <private/qobject_p.h>
 #include <private/qmetaobject_p.h>
+#if QT_CONFIG(qml_locale)
 #include <private/qqmllocale_p.h>
+#endif
 #include <private/qqmlbind_p.h>
 #include <private/qqmlconnections_p.h>
-#if QT_CONFIG(animation)
+#if QT_CONFIG(qml_animation)
 #include <private/qqmltimer_p.h>
 #endif
+#if QT_CONFIG(qml_list_model)
 #include <private/qqmllistmodel_p.h>
+#endif
 #include <private/qqmlplatform_p.h>
 #include <private/qquickpackage_p.h>
 #if QT_CONFIG(qml_delegate_model)
 #include <private/qqmldelegatemodel_p.h>
 #endif
 #include <private/qqmlobjectmodel_p.h>
+#if QT_CONFIG(qml_worker_script)
 #include <private/qquickworkerscript_p.h>
+#endif
 #include <private/qqmlinstantiator_p.h>
 #include <private/qqmlloggingcategory_p.h>
 
@@ -156,7 +161,7 @@ int qmlRegisterUncreatableMetaObject(const QMetaObject &staticMetaObject,
     \instantiates QObject
   \inqmlmodule QtQml
   \ingroup qml-utility-elements
-  \brief A basic QML type
+  \brief A basic QML type.
 
   The QtObject type is a non-visual element which contains only the
   objectName property.
@@ -217,30 +222,40 @@ void QQmlEnginePrivate::registerBaseTypes(const char *uri, int versionMajor, int
     qmlRegisterType<QObject>(uri,versionMajor,versionMinor,"QtObject");
     qmlRegisterType<QQmlBind>(uri, versionMajor, versionMinor,"Binding");
     qmlRegisterType<QQmlBind,8>(uri, versionMajor, (versionMinor < 8 ? 8 : versionMinor), "Binding"); //Only available in >=2.8
-    qmlRegisterType<QQmlConnections,1>(uri, versionMajor, (versionMinor < 3 ? 3 : versionMinor), "Connections"); //Only available in >=2.3
-    qmlRegisterType<QQmlConnections>(uri, versionMajor, versionMinor,"Connections");
-#if QT_CONFIG(animation)
+    qmlRegisterCustomType<QQmlConnections>(uri, versionMajor, 0, "Connections", new QQmlConnectionsParser);
+    if (!strcmp(uri, "QtQuick"))
+        qmlRegisterCustomType<QQmlConnections,1>(uri, versionMajor, 7, "Connections", new QQmlConnectionsParser); //Only available in QtQuick >=2.7
+    else
+        qmlRegisterCustomType<QQmlConnections,1>(uri, versionMajor, 3, "Connections", new QQmlConnectionsParser); //Only available in QtQml >=2.3
+#if QT_CONFIG(qml_animation)
     qmlRegisterType<QQmlTimer>(uri, versionMajor, versionMinor,"Timer");
 #endif
     qmlRegisterType<QQmlInstantiator>(uri, versionMajor, (versionMinor < 1 ? 1 : versionMinor), "Instantiator"); //Only available in >=2.1
-    qmlRegisterCustomType<QQmlConnections>(uri, versionMajor, versionMinor,"Connections", new QQmlConnectionsParser);
     qmlRegisterType<QQmlInstanceModel>();
-    qmlRegisterType<QQmlLoggingCategory>(uri, versionMajor, (versionMinor < 8 ? 8 : versionMinor), "LoggingCategory"); //Only available in >=2.8
+
+    qmlRegisterType<QQmlLoggingCategory>(uri, versionMajor, 8, "LoggingCategory"); //Only available in >=2.8
+    qmlRegisterType<QQmlLoggingCategory,1>(uri, versionMajor, 12, "LoggingCategory"); //Only available in >=2.12
 }
 
 
 // These QtQuick types' implementation resides in the QtQml module
 void QQmlEnginePrivate::registerQtQuick2Types(const char *uri, int versionMajor, int versionMinor)
 {
+#if QT_CONFIG(qml_list_model)
     qmlRegisterType<QQmlListElement>(uri, versionMajor, versionMinor, "ListElement"); // Now in QtQml.Models, here for compatibility
     qmlRegisterCustomType<QQmlListModel>(uri, versionMajor, versionMinor, "ListModel", new QQmlListModelParser); // Now in QtQml.Models, here for compatibility
+#endif
+#if QT_CONFIG(qml_worker_script)
     qmlRegisterType<QQuickWorkerScript>(uri, versionMajor, versionMinor, "WorkerScript");
+#endif
     qmlRegisterType<QQuickPackage>(uri, versionMajor, versionMinor, "Package");
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
 #if QT_CONFIG(qml_delegate_model)
     qmlRegisterType<QQmlDelegateModel>(uri, versionMajor, versionMinor, "VisualDataModel");
     qmlRegisterType<QQmlDelegateModelGroup>(uri, versionMajor, versionMinor, "VisualDataGroup");
 #endif
     qmlRegisterType<QQmlObjectModel>(uri, versionMajor, versionMinor, "VisualItemModel");
+#endif // < Qt 6
 }
 
 void QQmlEnginePrivate::defineQtQuick2Module()
@@ -250,7 +265,9 @@ void QQmlEnginePrivate::defineQtQuick2Module()
 
     // register the QtQuick2 types which are implemented in the QtQml module.
     registerQtQuick2Types("QtQuick",2,0);
+#if QT_CONFIG(qml_locale)
     qmlRegisterUncreatableType<QQmlLocale>("QtQuick", 2, 0, "Locale", QQmlEngine::tr("Locale cannot be instantiated.  Use Qt.locale()"));
+#endif
 
     // Auto-increment the import to stay in sync with ALL future QtQuick minor versions from 5.11 onward
     qmlRegisterModule("QtQuick", 2, QT_VERSION_MINOR);
@@ -676,7 +693,9 @@ QQmlEnginePrivate::QQmlEnginePrivate(QQmlEngine *e)
 #endif
   outputWarningsToMsgLog(true),
   cleanup(nullptr), erroredBindings(nullptr), inProgressCreations(0),
+#if QT_CONFIG(qml_worker_script)
   workerScriptEngine(nullptr),
+#endif
   activeObjectCreator(nullptr),
 #if QT_CONFIG(qml_network)
   networkAccessManager(nullptr), networkAccessManagerFactory(nullptr),
@@ -736,6 +755,9 @@ void QQmlPrivate::qdeclarativeelement_destructor(QObject *o)
             d->context = nullptr;
         }
 
+        if (d->outerContext && d->outerContext->contextObject == o)
+            d->outerContext->contextObject = nullptr;
+
         // Mark this object as in the process of deletion to
         // prevent it resolving in bindings
         QQmlData::markAsDeleted(o);
@@ -951,7 +973,9 @@ void QQmlEnginePrivate::init()
     if (baseModulesUninitialized) {
         qmlRegisterType<QQmlComponent>("QML", 1, 0, "Component"); // required for the Compiler.
         registerBaseTypes("QtQml", 2, 0); // import which provides language building blocks.
+#if QT_CONFIG(qml_locale)
         qmlRegisterUncreatableType<QQmlLocale>("QtQml", 2, 2, "Locale", QQmlEngine::tr("Locale cannot be instantiated.  Use Qt.locale()"));
+#endif
 
         // Auto-increment the import to stay in sync with ALL future QtQml minor versions from 5.11 onward
         qmlRegisterModule("QtQml", 2, QT_VERSION_MINOR);
@@ -974,6 +998,7 @@ void QQmlEnginePrivate::init()
     rootContext = new QQmlContext(q,true);
 }
 
+#if QT_CONFIG(qml_worker_script)
 QQuickWorkerScriptEngine *QQmlEnginePrivate::getWorkerScriptEngine()
 {
     Q_Q(QQmlEngine);
@@ -981,6 +1006,7 @@ QQuickWorkerScriptEngine *QQmlEnginePrivate::getWorkerScriptEngine()
         workerScriptEngine = new QQuickWorkerScriptEngine(q);
     return workerScriptEngine;
 }
+#endif
 
 /*!
   \class QQmlEngine
@@ -1010,8 +1036,6 @@ QQuickWorkerScriptEngine *QQmlEnginePrivate::getWorkerScriptEngine()
   In this case, the Text item will be created in the engine's
   \l {QQmlEngine::rootContext()}{root context}.
 
-  Note that the \l {Qt Quick 1} version is called QDeclarativeEngine.
-
   \sa QQmlComponent, QQmlContext, {QML Global Object}
 */
 
@@ -1280,9 +1304,7 @@ void QQmlEngine::addImageProvider(const QString &providerId, QQmlImageProviderBa
 }
 
 /*!
-  Returns the image provider set for \a providerId.
-
-  Returns the provider if it was found; otherwise returns 0.
+  Returns the image provider set for \a providerId if found; otherwise returns \nullptr.
 
   \sa QQuickImageProvider
 */
@@ -1364,6 +1386,71 @@ void QQmlEngine::setOutputWarningsToStandardError(bool enabled)
     d->outputWarningsToMsgLog = enabled;
 }
 
+/*!
+  \fn template<typename T> T QQmlEngine::singletonInstance(int qmlTypeId)
+
+  Returns the instance of a singleton type that was registered under \a qmlTypeId.
+
+  The template argument \e T may be either QJSValue or a pointer to a QObject-derived
+  type and depends on how the singleton was registered. If no instance of \e T has been
+  created yet, it is created now. If \a qmlTypeId does not represent a valid singleton
+  type, either a default constructed QJSValue or a \c nullptr is returned.
+
+  QObject* example:
+  \code
+  class MySingleton : public QObject {
+    Q_OBJECT
+    static int typeId;
+    // ...
+  };
+
+  // Register with QObject* callback
+  MySingleton::typeId = qmlRegisterSingletonType<MySingleton>(...);
+
+  // Retrieve as QObject*
+  QQmlEngine engine;
+  MySingleton* instance = engine.singletonInstance<MySingleton*>(MySingleton::typeId);
+  \endcode
+
+  QJSValue example:
+  \code
+  // Register with QJSValue callback
+  int typeId = qmlRegisterSingletonType(...);
+
+  // Retrieve as QJSValue
+  QQmlEngine engine;
+  QJSValue instance = engine.singletonInstance<QJSValue>(typeId);
+  \endcode
+
+  It is recommended to store the QML type id during registration, e.g. as a static member
+  in the singleton class. Otherwise, a costly lookup via qmlTypeId() has to be performed
+  at run-time.
+
+  \sa qmlRegisterSingletonType(), qmlTypeId()
+  \since 5.12
+*/
+template<>
+QJSValue QQmlEngine::singletonInstance<QJSValue>(int qmlTypeId)
+{
+    QQmlType type = QQmlMetaType::qmlType(qmlTypeId, QQmlMetaType::TypeIdCategory::QmlType);
+
+    if (!type.isValid() || !type.isSingleton())
+        return QJSValue();
+
+    QQmlType::SingletonInstanceInfo* info = type.singletonInstanceInfo();
+    info->init(this);
+
+    if (QObject* o = info->qobjectApi(this))
+        return this->newQObject(o);
+    else {
+        QJSValue value = info->scriptApi(this);
+        if (!value.isUndefined())
+            return value;
+    }
+
+    return QJSValue();
+}
+
 /*!
   Refreshes all binding expressions that use strings marked for translation.
 
@@ -1429,7 +1516,9 @@ void QQmlEngine::setContextForObject(QObject *object, QQmlContext *context)
     }
 
     QQmlContextData *contextData = QQmlContextData::get(context);
-    contextData->addObject(object);
+    Q_ASSERT(data->context == nullptr);
+    data->context = contextData;
+    contextData->addObject(data);
 }
 
 /*!
@@ -1699,7 +1788,7 @@ void QQmlData::NotifyList::layout()
     todo = nullptr;
 }
 
-void QQmlData::deferData(int objectIndex, QV4::CompiledData::CompilationUnit *compilationUnit, QQmlContextData *context)
+void QQmlData::deferData(int objectIndex, const QQmlRefPointer<QV4::CompiledData::CompilationUnit> &compilationUnit, QQmlContextData *context)
 {
     QQmlData::DeferredData *deferData = new QQmlData::DeferredData;
     deferData->deferredIdx = objectIndex;
@@ -1793,6 +1882,8 @@ void QQmlData::destroyed(QObject *object)
         nextContextObject->prevContextObject = prevContextObject;
     if (prevContextObject)
         *prevContextObject = nextContextObject;
+    else if (outerContext && outerContext->contextObjects == this)
+        outerContext->contextObjects = nextContextObject;
 
     QQmlAbstractBinding *binding = bindings;
     while (binding) {
@@ -2030,6 +2121,25 @@ void QQmlEnginePrivate::warning(QQmlEnginePrivate *engine, const QList<QQmlError
         dumpwarning(error);
 }
 
+QList<QQmlError> QQmlEnginePrivate::qmlErrorFromDiagnostics(const QString &fileName, const QList<DiagnosticMessage> &diagnosticMessages)
+{
+    QList<QQmlError> errors;
+    for (const DiagnosticMessage &m : diagnosticMessages) {
+        if (m.isWarning()) {
+            qWarning("%s:%d : %s", qPrintable(fileName), m.loc.startLine, qPrintable(m.message));
+            continue;
+        }
+
+        QQmlError error;
+        error.setUrl(QUrl(fileName));
+        error.setDescription(m.message);
+        error.setLine(m.loc.startLine);
+        error.setColumn(m.loc.startColumn);
+        errors << error;
+    }
+    return errors;
+}
+
 void QQmlEnginePrivate::cleanupScarceResources()
 {
     // iterate through the list and release them all.
@@ -2266,7 +2376,7 @@ QQmlMetaObject QQmlEnginePrivate::rawMetaObjectForType(int t) const
     Locker locker(this);
     auto iter = m_compositeTypes.constFind(t);
     if (iter != m_compositeTypes.cend()) {
-        return QQmlMetaObject((*iter)->rootPropertyCache());
+        return QQmlMetaObject((*iter)->rootPropertyCache().data());
     } else {
         QQmlType type = QQmlMetaType::qmlType(t);
         return QQmlMetaObject(type.baseMetaObject());
@@ -2278,7 +2388,7 @@ QQmlMetaObject QQmlEnginePrivate::metaObjectForType(int t) const
     Locker locker(this);
     auto iter = m_compositeTypes.constFind(t);
     if (iter != m_compositeTypes.cend()) {
-        return QQmlMetaObject((*iter)->rootPropertyCache());
+        return QQmlMetaObject((*iter)->rootPropertyCache().data());
     } else {
         QQmlType type = QQmlMetaType::qmlType(t);
         return QQmlMetaObject(type.metaObject());
@@ -2290,7 +2400,7 @@ QQmlPropertyCache *QQmlEnginePrivate::propertyCacheForType(int t)
     Locker locker(this);
     auto iter = m_compositeTypes.constFind(t);
     if (iter != m_compositeTypes.cend()) {
-        return (*iter)->rootPropertyCache();
+        return (*iter)->rootPropertyCache().data();
     } else {
         QQmlType type = QQmlMetaType::qmlType(t);
         locker.unlock();
@@ -2303,7 +2413,7 @@ QQmlPropertyCache *QQmlEnginePrivate::rawPropertyCacheForType(int t, int minorVe
     Locker locker(this);
     auto iter = m_compositeTypes.constFind(t);
     if (iter != m_compositeTypes.cend()) {
-        return (*iter)->rootPropertyCache();
+        return (*iter)->rootPropertyCache().data();
     } else {
         QQmlType type = QQmlMetaType::qmlType(t);
         locker.unlock();
