@@ -34,9 +34,12 @@
 #include <QDateTime>
 #include <QHashFunctions>
 #include <QSaveFile>
+#include <QScopedPointer>
+#include <QScopeGuard>
 
 #include <private/qqmlirbuilder_p.h>
 #include <private/qqmljsparser_p.h>
+#include <private/qqmljslexer_p.h>
 
 #include "resourcefilemapper.h"
 
@@ -66,6 +69,7 @@ struct Error
     QString message;
     void print();
     Error augment(const QString &contextErrorMessage) const;
+    void appendDiagnostics(const QString &inputFileName, const QList<QQmlJS::DiagnosticMessage> &diagnostics);
 };
 
 void Error::print()
@@ -95,6 +99,15 @@ QString diagnosticErrorMessage(const QString &fileName, const QQmlJS::Diagnostic
     return message;
 }
 
+void Error::appendDiagnostics(const QString &inputFileName, const QList<DiagnosticMessage> &diagnostics)
+{
+    for (const QQmlJS::DiagnosticMessage &parseError: diagnostics) {
+        if (!message.isEmpty())
+            message += QLatin1Char('\n');
+        message += diagnosticErrorMessage(inputFileName, parseError);
+    }
+}
+
 // Ensure that ListElement objects keep all property assignments in their string form
 static void annotateListElements(QmlIR::Document *document)
 {
@@ -147,11 +160,13 @@ static bool checkArgumentsObjectUseInSignalHandlers(const QmlIR::Document &doc,
                 if (compiledFunction->column > 0)
                     error->message += QString::number(compiledFunction->column) + QLatin1Char(':');
 
-                error->message += QLatin1String(" error: The use of the arguments object in signal handlers is\n"
-                                                "not supported when compiling qml files ahead of time, because it may be ambiguous if\n"
-                                                "any signal parameter is called \"arguments\". Unfortunately we cannot distinguish\n"
-                                                "between it being a parameter or the JavaScript arguments object at this point.\n"
-                                                "Consider renaming the parameter of the signal if applicable.");
+                error->message += QLatin1String(" error: The use of eval() or the use of the arguments object in signal handlers is\n"
+                                                "not supported when compiling qml files ahead of time. That is because it's ambiguous if \n"
+                                                "any signal parameter is called \"arguments\". Similarly the string passed to eval might use\n"
+                                                "\"arguments\". Unfortunately we cannot distinguish between it being a parameter or the\n"
+                                                "JavaScript arguments object at this point.\n"
+                                                "Consider renaming the parameter of the signal if applicable or moving the code into a\n"
+                                                "helper function.");
                 return false;
             }
         }
@@ -159,7 +174,7 @@ static bool checkArgumentsObjectUseInSignalHandlers(const QmlIR::Document &doc,
     return true;
 }
 
-using SaveFunction = std::function<bool (QV4::CompiledData::CompilationUnit *, QString *)>;
+using SaveFunction = std::function<bool (const QQmlRefPointer<QV4::CompiledData::CompilationUnit> &, QString *)>;
 
 static bool compileQmlFile(const QString &inputFileName, SaveFunction saveFunction, Error *error)
 {
@@ -182,11 +197,7 @@ static bool compileQmlFile(const QString &inputFileName, SaveFunction saveFuncti
     {
         QmlIR::IRBuilder irBuilder(illegalNames);
         if (!irBuilder.generateFromQml(sourceCode, inputFileName, &irDocument)) {
-            for (const QQmlJS::DiagnosticMessage &parseError: qAsConst(irBuilder.errors)) {
-                if (!error->message.isEmpty())
-                    error->message += QLatin1Char('\n');
-                error->message += diagnosticErrorMessage(inputFileName, parseError);
-            }
+            error->appendDiagnostics(inputFileName, irBuilder.errors);
             return false;
         }
     }
@@ -210,11 +221,7 @@ static bool compileQmlFile(const QString &inputFileName, SaveFunction saveFuncti
             const QVector<int> runtimeFunctionIndices = v4CodeGen.generateJSCodeForFunctionsAndBindings(functionsToCompile);
             QList<QQmlJS::DiagnosticMessage> jsErrors = v4CodeGen.errors();
             if (!jsErrors.isEmpty()) {
-                for (const QQmlJS::DiagnosticMessage &e: qAsConst(jsErrors)) {
-                    if (!error->message.isEmpty())
-                        error->message += QLatin1Char('\n');
-                    error->message += diagnosticErrorMessage(inputFileName, e);
-                }
+                error->appendDiagnostics(inputFileName, jsErrors);
                 return false;
             }
 
@@ -229,10 +236,10 @@ static bool compileQmlFile(const QString &inputFileName, SaveFunction saveFuncti
 
         QmlIR::QmlUnitGenerator generator;
         irDocument.javaScriptCompilationUnit = v4CodeGen.generateCompilationUnit(/*generate unit*/false);
-        QV4::CompiledData::Unit *unit = generator.generate(irDocument);
+        generator.generate(irDocument);
+        QV4::CompiledData::Unit *unit = const_cast<QV4::CompiledData::Unit*>(irDocument.javaScriptCompilationUnit->data);
         unit->flags |= QV4::CompiledData::Unit::StaticData;
         unit->flags |= QV4::CompiledData::Unit::PendingTypeCompilation;
-        irDocument.javaScriptCompilationUnit->data = unit;
 
         if (!saveFunction(irDocument.javaScriptCompilationUnit, &error->message))
             return false;
@@ -244,7 +251,8 @@ static bool compileQmlFile(const QString &inputFileName, SaveFunction saveFuncti
 
 static bool compileJSFile(const QString &inputFileName, const QString &inputFileUrl, SaveFunction saveFunction, Error *error)
 {
-    QmlIR::Document irDocument(/*debugMode*/false);
+    QQmlRefPointer<QV4::CompiledData::CompilationUnit> unit;
+    QScopedPointer<QV4::CompiledData::Unit, QScopedPointerPodDeleter> unitDataToFree;
 
     QString sourceCode;
     {
@@ -260,79 +268,83 @@ static bool compileJSFile(const QString &inputFileName, const QString &inputFile
         }
     }
 
-    QQmlJS::Engine *engine = &irDocument.jsParserEngine;
-    QmlIR::ScriptDirectivesCollector directivesCollector(&irDocument);
-    QQmlJS::Directives *oldDirs = engine->directives();
-    engine->setDirectives(&directivesCollector);
+    const bool isModule = inputFileName.endsWith(QLatin1String(".mjs"));
+    if (isModule) {
+        QList<QQmlJS::DiagnosticMessage> diagnostics;
+        // Precompiled files are relocatable and the final location will be set when loading.
+        QString url;
+        unit = QV4::ExecutionEngine::compileModule(/*debugMode*/false, url, sourceCode, QDateTime(), &diagnostics);
+        error->appendDiagnostics(inputFileName, diagnostics);
+        if (!unit)
+            return false;
+    } else {
+        QmlIR::Document irDocument(/*debugMode*/false);
 
-    QQmlJS::AST::Program *program = nullptr;
+        QQmlJS::Engine *engine = &irDocument.jsParserEngine;
+        QmlIR::ScriptDirectivesCollector directivesCollector(&irDocument);
+        QQmlJS::Directives *oldDirs = engine->directives();
+        engine->setDirectives(&directivesCollector);
+        auto directivesGuard = qScopeGuard([engine, oldDirs]{
+            engine->setDirectives(oldDirs);
+        });
 
-    {
-        QQmlJS::Lexer lexer(engine);
-         lexer.setCode(sourceCode, /*line*/1, /*parseAsBinding*/false);
-         QQmlJS::Parser parser(engine);
+        QQmlJS::AST::Program *program = nullptr;
 
-         bool parsed = parser.parseProgram();
+        {
+            QQmlJS::Lexer lexer(engine);
+            lexer.setCode(sourceCode, /*line*/1, /*parseAsBinding*/false);
+            QQmlJS::Parser parser(engine);
 
-         for (const QQmlJS::DiagnosticMessage &parseError: parser.diagnosticMessages()) {
-             if (!error->message.isEmpty())
-                 error->message += QLatin1Char('\n');
-             error->message += diagnosticErrorMessage(inputFileName, parseError);
-         }
+            bool parsed = parser.parseProgram();
 
-         if (!parsed) {
-             engine->setDirectives(oldDirs);
-             return false;
-         }
+            error->appendDiagnostics(inputFileName, parser.diagnosticMessages());
 
-         program = QQmlJS::AST::cast<QQmlJS::AST::Program*>(parser.rootNode());
-         if (!program) {
-             lexer.setCode(QStringLiteral("undefined;"), 1, false);
-             parsed = parser.parseProgram();
-             Q_ASSERT(parsed);
-             program = QQmlJS::AST::cast<QQmlJS::AST::Program*>(parser.rootNode());
-             Q_ASSERT(program);
-         }
-    }
+            if (!parsed)
+                return false;
 
-    {
-        QmlIR::JSCodeGen v4CodeGen(irDocument.code, &irDocument.jsGenerator,
-                                   &irDocument.jsModule, &irDocument.jsParserEngine,
-                                   irDocument.program, /*import cache*/nullptr,
-                                   &irDocument.jsGenerator.stringTable, illegalNames);
-        v4CodeGen.setUseFastLookups(false); // Disable lookups in non-standalone (aka QML) mode
-        v4CodeGen.generateFromProgram(inputFileName, inputFileUrl, sourceCode, program,
-                                      &irDocument.jsModule, QV4::Compiler::GlobalCode);
-        QList<QQmlJS::DiagnosticMessage> jsErrors = v4CodeGen.errors();
-        if (!jsErrors.isEmpty()) {
-            for (const QQmlJS::DiagnosticMessage &e: qAsConst(jsErrors)) {
-                if (!error->message.isEmpty())
-                    error->message += QLatin1Char('\n');
-                error->message += diagnosticErrorMessage(inputFileName, e);
+            program = QQmlJS::AST::cast<QQmlJS::AST::Program*>(parser.rootNode());
+            if (!program) {
+                lexer.setCode(QStringLiteral("undefined;"), 1, false);
+                parsed = parser.parseProgram();
+                Q_ASSERT(parsed);
+                program = QQmlJS::AST::cast<QQmlJS::AST::Program*>(parser.rootNode());
+                Q_ASSERT(program);
             }
-            engine->setDirectives(oldDirs);
-            return false;
         }
 
-        QmlIR::QmlUnitGenerator generator;
+        {
+            QmlIR::JSCodeGen v4CodeGen(irDocument.code, &irDocument.jsGenerator,
+                                       &irDocument.jsModule, &irDocument.jsParserEngine,
+                                       irDocument.program, /*import cache*/nullptr,
+                                       &irDocument.jsGenerator.stringTable, illegalNames);
+            v4CodeGen.setUseFastLookups(false); // Disable lookups in non-standalone (aka QML) mode
+            v4CodeGen.generateFromProgram(inputFileName, inputFileUrl, sourceCode, program,
+                                          &irDocument.jsModule, QV4::Compiler::ContextType::ScriptImportedByQML);
+            QList<QQmlJS::DiagnosticMessage> jsErrors = v4CodeGen.errors();
+            if (!jsErrors.isEmpty()) {
+                error->appendDiagnostics(inputFileName, jsErrors);
+                return false;
+            }
 
-        irDocument.javaScriptCompilationUnit = v4CodeGen.generateCompilationUnit(/*generate unit*/false);
-        QV4::CompiledData::Unit *unit = generator.generate(irDocument);
-        unit->flags |= QV4::CompiledData::Unit::StaticData;
-        irDocument.javaScriptCompilationUnit->data = unit;
+            // Precompiled files are relocatable and the final location will be set when loading.
+            irDocument.jsModule.fileName.clear();
+            irDocument.jsModule.finalUrl.clear();
 
-        if (!saveFunction(irDocument.javaScriptCompilationUnit, &error->message)) {
-            engine->setDirectives(oldDirs);
-            return false;
+            irDocument.javaScriptCompilationUnit = v4CodeGen.generateCompilationUnit(/*generate unit*/false);
+            QmlIR::QmlUnitGenerator generator;
+            generator.generate(irDocument);
+            QV4::CompiledData::Unit *unitData = const_cast<QV4::CompiledData::Unit*>(irDocument.javaScriptCompilationUnit->data);
+            unitData->flags |= QV4::CompiledData::Unit::StaticData;
+            unitDataToFree.reset(unitData);
+            unit = irDocument.javaScriptCompilationUnit;
         }
-
-        free(unit);
     }
-    engine->setDirectives(oldDirs);
-    return true;
+
+    return saveFunction(unit, &error->message);
 }
 
-static bool saveUnitAsCpp(const QString &inputFileName, const QString &outputFileName, QV4::CompiledData::CompilationUnit *unit, QString *errorString)
+static bool saveUnitAsCpp(const QString &inputFileName, const QString &outputFileName,
+                          const QQmlRefPointer<QV4::CompiledData::CompilationUnit> &unit, QString *errorString)
 {
     QSaveFile f(outputFileName);
     if (!f.open(QIODevice::WriteOnly | QIODevice::Truncate)) {
@@ -432,9 +444,6 @@ int main(int argc, char **argv)
     QCommandLineOption outputFileOption(QStringLiteral("o"), QCoreApplication::translate("main", "Output file name"), QCoreApplication::translate("main", "file name"));
     parser.addOption(outputFileOption);
 
-    QCommandLineOption checkIfSupportedOption(QStringLiteral("check-if-supported"), QCoreApplication::translate("main", "Check if cache generate is supported on the specified target architecture"));
-    parser.addOption(checkIfSupportedOption);
-
     parser.addPositionalArgument(QStringLiteral("[qml file]"),
             QStringLiteral("QML source file to generate cache for."));
 
@@ -523,12 +532,12 @@ int main(int argc, char **argv)
 
         inputFileUrl = QStringLiteral("qrc://") + inputResourcePath;
 
-        saveFunction = [inputResourcePath, outputFileName](QV4::CompiledData::CompilationUnit *unit, QString *errorString) {
+        saveFunction = [inputResourcePath, outputFileName](const QQmlRefPointer<QV4::CompiledData::CompilationUnit> &unit, QString *errorString) {
             return saveUnitAsCpp(inputResourcePath, outputFileName, unit, errorString);
         };
 
     } else {
-        saveFunction = [outputFileName](QV4::CompiledData::CompilationUnit *unit, QString *errorString) {
+        saveFunction = [outputFileName](const QQmlRefPointer<QV4::CompiledData::CompilationUnit> &unit, QString *errorString) {
             return unit->saveToDisk(outputFileName, errorString);
         };
     }
@@ -542,7 +551,7 @@ int main(int argc, char **argv)
             error.augment(QLatin1String("Error compiling qml file: ")).print();
             return EXIT_FAILURE;
         }
-    } else if (inputFile.endsWith(QLatin1String(".js"))) {
+    } else if (inputFile.endsWith(QLatin1String(".js")) || inputFile.endsWith(QLatin1String(".mjs"))) {
         Error error;
         if (!compileJSFile(inputFile, inputFileUrl, saveFunction, &error)) {
             error.augment(QLatin1String("Error compiling js file: ")).print();
